// Tetris.cpp: implementation of the CTetris class.
// A device/graphic library independent Tetris class
// DO NOT add any graphics-specific (i.e. CDX) code in this class
// Written by Jimb Esser
// Original Tetris game concept Alexy Pajitnov
//////////////////////////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "Tetris.h"
#include "PieceLib.h"
#include <assert.h>
#include <stdlib.h>

#include "myrand.h"

// Piece constants
int pieces[] =		   {0, 0, 0, 0,
						0, 1, 1, 0,
						0, 1, 1, 0,
						0, 0, 0, 0,
/*t pc_l[] */			0, 1, 0, 0,
						0, 1, 0, 0,
						0, 1, 1, 0,
						0, 0, 0, 0,
/*t pc_j[] */			0, 0, 1, 0,
						0, 0, 1, 0,
						0, 1, 1, 0,
						0, 0, 0, 0,
/* pc_s[] */			0, 1, 0, 0,
						0, 1, 1, 0,
						0, 0, 1, 0,
						0, 0, 0, 0,
/* pc_z[] */			0, 0, 1, 0,
						0, 1, 1, 0,
						0, 1, 0, 0,
						0, 0, 0, 0,
/* pc_t[] */			0, 1, 0, 0,
						0, 1, 1, 0,
						0, 1, 0, 0,
						0, 0, 0, 0,
/* pc_line[] */			0, 1, 0, 0,
						0, 1, 0, 0,
						0, 1, 0, 0,
						0, 1, 0, 0};


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CTetris::CTetris()
{
	pieceLib=NULL;
	bJustSwapped=false;
	Reset();
}

CTetris::~CTetris()
{
	if (pieceLib!=NULL) {
		delete pieceLib;
		pieceLib=NULL;
	}
}



bool CTetris::Clash()
{
	int x, y;
	for (x=0; x<P_WIDTH; x++)
		for (y=0; y<P_HEIGHT; y++) {
			if ((x+px>=0) && (x+px<TET_WIDTH) &&
				(y+py>=0) && (y+py<TET_HEIGHT)) {
				// In bounds
				if (piece[x][y] && board[x+px][y+py])
					// And it overlaps a piece of the board
					return true;
			} else { // out of bounds
				if (piece[x][y])
					// And part of the piece is out of bounds
					return true;
			}
		}
	return false;
}

void CTetris::SetPiece()
{
	int x, y;
	for (x=0; x<P_WIDTH; x++)
		for (y=0; y<P_HEIGHT; y++) {
			if ((x+px>=0) && (x+px<TET_WIDTH) &&
				(y+py>=0) && (y+py<TET_HEIGHT))
				if (piece[x][y])
					board[x+px][y+py] = liPieceID;
		}
	CheckGoldSilver();
	CheckMadeLine();
	NewPiece();
}

void CTetris::goDown()
{
	if (!bDead) {
		py++;
		if (Clash()) {
			py--;
			SetPiece();
		}
		dwLastAction=0;
	}
}

void CTetris::goRight()
{
	if (!bDead) {
		px++;
		if (Clash()) {
			px--;
		}
	}
}

void CTetris::goLeft()
{
	if (!bDead) {
		px--;
		if (Clash()) {
			px++;
		}
	}
}

void CTetris::goDrop()
{
	if (!bDead) {
		while (!Clash()) {
			py++;
		}
		py--;
		SetPiece();
		dwLastAction=0;
	}
}

void CopyPiece(piece_t &dest, piece_t &src)
{
	for (int x=0; x<P_WIDTH; x++)
		for (int y=0; y<P_HEIGHT; y++)
			dest[x][y] = src[x][y];
}

void CTetris::RotateCW()
{
	piece_t oldPiece;
	CopyPiece(oldPiece, piece);
	int x, y;
	for (x=0; x<P_WIDTH; x++)
		for (y=0; y<P_HEIGHT; y++)
			piece[x][y] = oldPiece[y][P_HEIGHT-1-x];
	if (Clash()) {
		CopyPiece(piece, oldPiece);
	}
}

void CTetris::RotateCCW()
{
	piece_t oldPiece;
	CopyPiece(oldPiece, piece);
	int x, y;
	for (x=0; x<P_WIDTH; x++)
		for (y=0; y<P_HEIGHT; y++)
			piece[x][y] = oldPiece[P_WIDTH-1-y][x];
	if (Clash()) {
		CopyPiece(piece, oldPiece);
	}
}

void CTetris::NewPiece()
{
	if (!bDead) {
		CopyPiece(piece, nextpiece);
		int iPiece = myrand(iSeed++); //rand()%NUMPEICES;
		liPieceID = liNextPieceID;
		liNextPieceID = pieceLib->getNewID(iPiece+1);
		for (int i=0; i<P_WIDTH*P_HEIGHT; i++) {
			nextpiece[i%P_WIDTH][i/P_WIDTH] = pieces[iPiece*P_WIDTH*P_HEIGHT + i] * (iPiece+1);
			// The * (iPiece+1) sets a number to the squares of the piece, instead of everything being just "1"
		}
		px=(TET_WIDTH - P_WIDTH)/2;
		py=0;

		if (Clash())
			bDead = true;
		bJustSwapped=false;
	}
}

board_t *CTetris::getDrawnBoard()
{
	int x, y;
	for (x=0; x<TET_WIDTH; x++)
		for (y=0; y<TET_HEIGHT; y++)
			drawnBoard[x][y] = pieceLib->getType(board[x][y]);
	for (x=0; x<P_WIDTH; x++)
		for (y=0; y<P_HEIGHT; y++) {
			if ((x+px>=0) && (x+px<TET_WIDTH) &&
				(y+py>=0) && (y+py<TET_HEIGHT))
				if (piece[x][y])
					drawnBoard[x+px][y+py] = piece[x][y];
		}
	return &drawnBoard;
}

void CTetris::Reset()
{
	int x, y;
	if (pieceLib==NULL) {
		pieceLib = new CPieceLib();
	}
	pieceLib->Reset();
	LOOPBOARD(x, y) {
		board[x][y] = 0;
	}
	bDead = false;
	iSeed=0;
	NewPiece(); // Do this twice so that both a new piece and the nextpiece are created
	NewPiece();
	swapToSaved(); // Do this so we have something in our saved piece
	NewPiece();
	iScore = 0;
	iLines = 0;
	dwLastAction=0;
}


void CTetris::CheckGoldSilver()
{
	int x, x2, y, y2, i;
	long id;
	long ids[4];
	int numIds;
	for (y=0; y<=TET_HEIGHT-4; y++) {
		for (x=0; x<=TET_WIDTH-4; x++) {
			if (board[x][y]) {
				numIds=0;
				for (x2=0; x2<4 && numIds<=4;  x2++) {
					for (y2=0; y2<4; y2++) {
						id = board[x+x2][y+y2];
						if (id==0) {
							numIds=5; // Break
						} else if ((i=pieceLib->getType(id))==TYPE_SILVER ||
							i==TYPE_GOLD) {
							numIds=5; // Break out, already has a gold or silver
						} else {
							for (i=0; i<numIds; i++) {
								if (ids[i]==id) {
									i=5; // we already have this ID
								}
							}
							if (i==numIds) {
								// We have a new ID
								if (numIds>=4) {
									// This does not qualify for a gold/silver, more than 4 ids
								} else {
									ids[numIds]=id;
								}
								numIds++; // inc num of ids
							}
						}
					}
				}
//				assert(numIds>=4);
				if (numIds==4) {
					// Gold or silver!
					bool bGold = true;
					int iType = pieceLib->getType(ids[0]);
					for (i=1; i<4; i++) {
						if (pieceLib->getType(ids[i])!=iType) {
							bGold=false;
						}
					}
					long liNewID = pieceLib->getNewID(bGold?TYPE_GOLD:TYPE_SILVER);
					for (x2=0; x2<4;  x2++) {
						for (y2=0; y2<4; y2++) {
							board[x+x2][y+y2]=liNewID;
						}
					}
					x+=3; // Skip past this horizontally
				}
			}
		}
	}


}

void CTetris::CheckMadeLine()
{
	int iNumActualLines = 0;
	int iNumLines = 0; // Just gold and silver bonuses
	int x, y, x2, y2, iGP, iSP;
	long iMult, iTile;
	for (y=TET_HEIGHT-1; y>=0; y--) {
		iMult=1;
		iGP=0;
		iSP=0;
		for (x=0; x<TET_WIDTH; x++) {
			iTile = board[x][y];
			iMult*=iTile;
			if (pieceLib->getType(iTile)==TYPE_GOLD) {
				iGP++;
			} else if (pieceLib->getType(iTile)==TYPE_SILVER) {
				iSP++;
			}
		}
		if (iMult!=0) { // If there was any blank spot (a zero) iMult would be = 0
			// Made a line
			iNumActualLines++;
			if (iGP==4) {
				iNumLines+=10;
				if (iSP==4) {
					iNumLines+=4;
				}
			} else if (iGP==8) {
				iNumLines+=20;
			} else if (iSP==4) {
				iNumLines+=4;
			} else if (iSP==8) {
				iNumLines+=8;
			}
			for (y2=y; y2>0; y2--)
				for (x2=0; x2<TET_WIDTH; x2++)
					board[x2][y2] = board[x2][y2-1];
			for (x2=0; x2<TET_WIDTH; x2++)
				board[x2][0] = 0;
			y++; // Need to recheck this line, since it now has new stuff in it.
		}
	}
	// Increment the score
	int iTotalMultiplier=1; // Based on number of lines (2^(lines-1))
	switch (iNumActualLines) {
	case 4:
		iTotalMultiplier*=2;
	case 3:
		iTotalMultiplier*=2;
	case 2:
		iTotalMultiplier*=2;
		break;
	case 0:
		iTotalMultiplier=0;
	}
	iScore += iTotalMultiplier + iNumLines;
	iLines += iNumActualLines;
}

void CTetris::swapToSaved() {

	if (!bDead && !bJustSwapped) { 
		bJustSwapped=true;
		piece_t temp;
		CopyPiece(temp, piece);
		CopyPiece(piece, savedpiece);
		CopyPiece(savedpiece, temp);
		long t=liPieceID;
		liPieceID = liSavedPieceID;
		liSavedPieceID = t;

		px=(TET_WIDTH - P_WIDTH)/2;
		py=0;
	}
}

void CTetris::Update(DWORD dwTime)
{
	if (dwLastAction==0) {
		// First time calling the Update function since a reset
		dwLastAction=dwTime;
	} else {
		if (dwTime > dwLastAction + MS_PER_DROP) {
			dwLastAction = dwTime;
			goDown();
		}
	}
}
